@@ -1,6 +1,7 @@
 /* This file is part of q-tools, a collection of performance tools
-   Copyright (c) 2003-2004 Hewlett-Packard Development Company, L.P.
+   Copyright (c) 2003-2006 Hewlett-Packard Development Company, L.P.
    Contributed by David Mosberger-Tang <davidm@hpl.hp.com>
+   Montecito support by Stephane Eranian <eranian@hpl.hp.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -31,18 +32,6 @@
 #include <syscall.h>
 #include <unistd.h>
 
-#ifdef HAVE_EXPLICIT_PERFMON3
-# include <perfmon3/perfmon.h>
-# include <perfmon3/perfmon_default_smpl.h>
-# include <perfmon3/pfmlib.h>
-# include <perfmon3/pfmlib_itanium2.h>
-#else
-# include <perfmon/perfmon.h>
-# include <perfmon/perfmon_default_smpl.h>
-# include <perfmon/pfmlib.h>
-# include <perfmon/pfmlib_itanium2.h>
-#endif
-
 #include <sys/poll.h>
 #include <sys/time.h>
 #include <sys/types.h>
@@ -54,18 +43,6 @@
 #include "call-counts.h"
 #include "time-hist.h"
 
-#define NELEMS(a)		((sizeof (a) / sizeof (a[0])))
-#define REG_MASK(regnum)	(1UL << (regnum))
-
-#define KB			(1 << 10)
-#define SAMPLING_BUFFER_SIZE	(256*KB)
-#define BTB_REGS_MASK		(  REG_MASK (8)  | REG_MASK (9)		\
-				 | REG_MASK (10) | REG_MASK (11)	\
-				 | REG_MASK (12) | REG_MASK (13)	\
-				 | REG_MASK (14) | REG_MASK (15)	\
-				 | REG_MASK (16))
-#define BTB_LEN			4	/* McKinley-specific */
-
 /* It is currently not possible to add events incrementally.  Because
    of that, we need a gross hack in setup_code_sampling() so libpfm
    knows that PMD4 is already taken for BRANCH_EVENT when we dispatch
@@ -103,27 +80,6 @@ static struct option long_opts[] =
 const char *prog_name;
 const char **prog_argv;
 
-struct mapping
-  {
-    struct mapping *next;	/* must be first member in this structure! */
-    unsigned long start;
-    unsigned long size;
-    unsigned long offset;
-    const char *path;
-  };
-
-struct addr_space
-  {
-    struct addr_space *next;
-    pid_t pid;
-    pid_t tgid;		/* thread-group id (defaults to same as pid) */
-    const char *name;
-    const char **argv;
-    struct time_hist *hist;
-    struct call_count *cc;
-    struct mapping *mappings;
-  };
-
 static int verbose;
 static pid_t *child_pid;
 static struct addr_space *addr_space_list;
@@ -135,21 +91,34 @@ static int poll_timeout = MIN_POLL_TIMEOUT;
 static int num_cpus, my_cpu = -1;
 /* PMD register number that's couting BRANCH_EVENT: */
 static int branch_event_pmd = -1;
+static pfarg_reg_t *branch_event_pd, *cs_pd;
 /* ditto for code-sampling: */
 static int cs_event_pmd = -1;
 static int duration;
-static double code_sample_rate = 1000.0;
-static double call_sample_rate = 1000.0;
-static const char *cs_event_name = "CPU_CYCLES";
+static char *cs_event_name;
 static int num_pds;
 static pfarg_reg_t pd[PFMLIB_MAX_PMDS];
-static pfarg_reg_t *branch_event_pd, *cs_pd;
 static struct timeval now, last_time;
 static sig_atomic_t done;
-static unsigned long cycle_frequency;
 static unsigned long cs_interval_sum, cs_num_samples;
 static pfm_default_smpl_hdr_t *sampling_buffer;
 static int sampling_fd;
+static struct qsys_btb_support *qsys;
+
+unsigned int cs_event_code, cs_cycle_code;
+unsigned long cycle_frequency;
+double code_sample_rate = 1000.0;
+double call_sample_rate = 1000.0;
+
+extern struct qsys_btb_support qsys_itanium2, qsys_montecito;
+
+static struct qsys_btb_support *qsys_support_list[]={
+#ifdef __ia64__
+	&qsys_itanium2,
+	&qsys_montecito,
+#endif
+	NULL
+};
 
 /* Find last bit set.  Returns -1 for 0.  */
 static inline long
@@ -423,7 +392,7 @@ update_mappings (struct addr_space *as)
   as->mappings = list;
 }
 
-static void
+void
 check_addr_space_mapping (struct addr_space *as, unsigned long addr)
 {
   struct mapping *m, *prev = (struct mapping *) &as->mappings;
@@ -448,189 +417,6 @@ check_addr_space_mapping (struct addr_space *as, unsigned long addr)
   update_mappings (as);
 }
 
-static inline uint64_t
-get_btb_ip (int regnum, pfm_ita2_pmd_reg_t reg, pfm_ita2_pmd_reg_t pmd16)
-{
-  /* We can ignore the slot-number here because functions must start
-     and end at a bundle-boundary.  */
-  uint64_t ip = (reg.pmd8_15_ita2_reg.btb_addr << 4);
-
-  if (pmd16.pmd_val & (1UL << (4 + 4*regnum)))
-    ip += 0x10;
-
-  return ip;
-}
-
-static char *
-process_btb_sample (struct addr_space *as,
-		    pfm_default_smpl_entry_t *ent, char *pos)
-{
-  pfm_ita2_pmd_reg_t pmd16;
-  uint64_t j, src_ip, dst_ip, last, interval;
-  pfm_ita2_pmd_reg_t *reg, src, dst;
-  unsigned long weight;
-  int n;
-
-  reg = (pfm_ita2_pmd_reg_t *) pos;
-
-  interval = -ent->last_reset_val;
-
-  pmd16 = reg[8];
-
-  last = pmd16.pmd16_ita2_reg.btbi_bbi;
-  j = pmd16.pmd16_ita2_reg.btbi_full ? pmd16.pmd16_ita2_reg.btbi_bbi : 0;
-
-#if 0
-  printf ("first=%lu, last=%lu, pmd16=%lx, interval=%ld\n",
-	  j, last, pmd16.pmd_val, interval);
-  {
-    int k;
-    for (k = 0; k < 9; ++k)
-      printf ("\treg[%d] = %016lx\n", k, reg[k].pmd_val);
-  }
-#endif
-
-  /* count the number of branches recorded in the BTB: */
-  n = 0;
-  do
-    {
-      src = reg[j];
-      dst = reg[(j + 1) & 7];
-
-      if (src.pmd8_15_ds_ita2_reg.btb_b)
-	{
-	  ++n;
-	  if (!dst.pmd8_15_ds_ita2_reg.btb_b)
-	    j = (j + 1) & 7;
-	  if (j == last)
-	    break;
-	}
-      j = (j + 1) & 7;
-    }
-  while (j != last);
-
-  j =  pmd16.pmd16_ita2_reg.btbi_full ? pmd16.pmd16_ita2_reg.btbi_bbi : 0;
-
-  if (n > 0)
-    {
-#if 0
-      n = 1;
-#endif
-      weight = (2*interval + n) / (2*n);
-
-      while (n-- > 0)
-	{
-	  src = reg[j];
-	  dst = reg[(j + 1) & 7];
-
-	  /* Ensure "src" recorded a branch instruction (btb_b set to
-	     0), not a branch target.  */
-	  if (src.pmd8_15_ds_ita2_reg.btb_b)
-	    {
-	      src_ip = get_btb_ip (j, src, pmd16);
-	      dst_ip = get_btb_ip ((j + 1) & 7, dst, pmd16);
-
-	      /* Note that we're capturing br.ret, so a return from
-		 SRC to DST gets count as a call from DST to
-		 SRC...  */
-	      call_count_add (as->cc, dst_ip, src_ip, weight);
-	      check_addr_space_mapping (as, src_ip);
-	      check_addr_space_mapping (as, dst_ip);
-
-	      /* If destination address was recorded by a branch
-		 target entry (btb_b set to 0), skip over that entry
-		 as well.  */
-	      if (!dst.pmd8_15_ds_ita2_reg.btb_b)
-		j = (j + 1) & 7;
-	    }
-	  j = (j + 1) & 7;
-	}
-    }
-  return (char *) &reg[9];	/* XXX must be 16-byte aligned??? */
-}
-
-inline static void
-dump_btb (pfm_ita2_pmd_reg_t *pmd)
-{
-  unsigned long reg, i, bbi, num_regs;
-  pfm_ita2_pmd_reg_t pmd16;
-
-  pmd16 = pmd[8];
-  bbi = pmd16.pmd16_ita2_reg.btbi_bbi;
-
-  if (pmd16.pmd16_ita2_reg.btbi_full)
-    num_regs = 8;
-  else
-    num_regs = bbi + 1;
-  i = (bbi + 7) % 8;
-  printf ("--- bbi=%lu\n", bbi);
-  for (reg = 0; reg < num_regs; ++reg)
-    {
-      printf ("BTB[%lu] = %016lx\n", i, *(unsigned long *)&pmd[i]);
-      i = (i + 7) % 8;
-    }
-}
-
-static uint64_t
-get_ip_from_btb (pfm_default_smpl_entry_t *ent, char **posp)
-{
-  unsigned long reg, num_regs, i, j, bbi, src_addr, dst_addr;
-  pfm_ita2_pmd_reg_t *pmd, src, dst, pmd16;
-
-  pmd = (pfm_ita2_pmd_reg_t *) *posp;
-  *posp = (char *) &pmd[9];
-
-  /* find the most recently recorded branch in the BTB: */
-  pmd16 = pmd[8];
-  bbi = pmd16.pmd16_ita2_reg.btbi_bbi;
-
-  if (pmd16.pmd16_ita2_reg.btbi_full)
-    num_regs = 8;
-  else
-    num_regs = bbi + 1;
-  i = (bbi + 7) % 8;
-  for (reg = 0; reg < num_regs; ++reg)
-    {
-      src = pmd[i];
-      if (src.pmd8_15_ita2_reg.btb_b)
-	{
-	  /* Found the register recording the most-recent branch
-	     instruction.  */
-
-	  /* As a special case, if the most recent branch was a taken
-	     branch from the last slot in a bundle to the next bundle,
-	     we treat it as an "rfi" marker and ignore it.  */
-	  if (src.pmd8_15_ita2_reg.btb_slot == 2 && reg > 0)
-	    {
-	      j = (i + 1) & 7;
-	      dst = pmd[j];
-	      src_addr = get_btb_ip (i, src, pmd16);
-	      dst_addr = get_btb_ip (j, dst, pmd16);
-	      if (dst_addr - src_addr == 0x10)
-		return ent->ip | 0xd;
-	    }
-
-	  if (src.pmd8_15_ita2_reg.btb_slot != 3 && reg > 0)
-	    {
-	      /* the branch was taken and we have a record of the
-		 branch destination; record that instead */
-	      i = (i + 1) & 7;
-	      src = pmd[i];
-	    }
-	  break;
-	}
-      i = (i + 7) % 8;
-    }
-  if (reg >= num_regs)
-    {
-      /* BTB didn't record any branches!  Use IP from interrupt
-	 instead and mark bits 0-3 with "0xc" so we can distinguish
-	 the sample from BTB-samples. */
-      return ent->ip | 0xc;	/* or in with special marker... */
-    }
-  return get_btb_ip (i, src, pmd16);
-}
-
 static char *
 process_code_sample (struct addr_space *as,
 		     pfm_default_smpl_entry_t *ent, char *pos)
@@ -642,7 +428,7 @@ process_code_sample (struct addr_space *as,
   ++cs_num_samples;
 
   if (monitor_interruptions)
-    ip = get_ip_from_btb (ent, &pos);
+    ip = qsys->get_ip_from_btb (ent, &pos);
   else
     ip = ent->ip;
 
@@ -730,27 +516,23 @@ process_pmu_samples (void)
   int need_adjustment;
   pid_t pid = -1;
   double delta;
-  char *pos;
+  char *pos, *old_pos;
 
   last_time = now;
   gettimeofday (&now, NULL);
 
-#if 0
-  printf ("%s(count=%lu)\n", __FUNCTION__, hdr->hdr_count);
-#endif
 
-  if (hdr->hdr_version != PFM_DEFAULT_SMPL_VERSION)
-    panic ("pfmlib v%u.%u sampling format is not supported\n",
-	   PFM_VERSION_MAJOR (hdr->hdr_version),
-	   PFM_VERSION_MINOR (hdr->hdr_version));
-
-  pos = (char *) (hdr + 1);
+    pos = (char *) (hdr + 1);
 
   for (i = 0; i < hdr->hdr_count; ++i)
     {
       ent = (pfm_default_smpl_entry_t *) pos;
-      pos = (char *) (ent + 1);
+      pos = old_pos = (char *) (ent + 1);
 
+#if 0
+  printf ("%s(count=%lu) btb_samples=%lu code_samples=%lu\n",
+	  __FUNCTION__, hdr->hdr_count, num_btb_samples, num_code_samples);
+#endif
       if (pid != ent->pid)
 	{
 	  pid = ent->pid;
@@ -760,7 +542,7 @@ process_pmu_samples (void)
       if (ent->ovfl_pmd == branch_event_pmd)
 	{
 	  ++num_btb_samples;
-	  pos = process_btb_sample (as, ent, pos);
+	  pos = qsys->process_btb_sample(as, ent, pos);
 	}
       else if (ent->ovfl_pmd == cs_event_pmd)
 	{
@@ -768,7 +550,8 @@ process_pmu_samples (void)
 	  pos = process_code_sample (as, ent, pos);
 	}
       else
-	panic ("Unknown PMD register %u overflowed\n", ent->ovfl_pmd);
+	panic ("%lu Unknown PMD register %u overflowed (%u)\n", i, ent->ovfl_pmd, branch_event_pmd);
+
     }
 
   delta = ((now.tv_sec + 1e-6 * now.tv_usec)
@@ -906,7 +689,7 @@ write_profile (struct addr_space *as, const char *kallsyms)
 	   "               (q:event-name . \"%s\")\n"
 	   "               (q:x-unit-label . \"address\")\n",
 	   hist_file_name, cs_event_name);
-  if (strcmp (cs_event_name, "CPU_CYCLES") == 0)
+  if (cs_event_code == cs_cycle_code)
     fprintf (info,
 	     "               (q:y-unit-label . \"seconds\")\n"
 	     "               (q:y-unit-conversion-factor . %g)\n",
@@ -990,93 +773,10 @@ write_profile (struct addr_space *as, const char *kallsyms)
   if (edge)
     q_close (edge);
 }
+	
 
-static void
-setup_call_count_sampling (pfarg_reg_t *pc, int *num_pcsp,
-			   pfarg_reg_t *pd, int *num_pdsp,
-			   unsigned int privilege_level_mask)
-{
-  int j, ret, type, num_pcs = *num_pcsp, num_pds = *num_pdsp;
-  pfmlib_ita2_input_param_t i2param;
-  pfmlib_input_param_t iparam;
-  pfmlib_output_param_t oparam;
-  unsigned long reset_value;
-  unsigned long i;
 
-  pfm_get_pmu_type (&type);
-  if (type != PFMLIB_ITANIUM2_PMU)
-    {
-      char model[80];
-      pfm_get_pmu_name (model, sizeof (model));
-      fprintf (stderr, "Call-count sampling not supported by %s PMU; "
-	       "disabling it.\n", model);
-      return;		/* not a fatal error... */
-    }
 
-  memset (&iparam, 0, sizeof (&iparam));
-  memset (&i2param, 0, sizeof (i2param));
-
-  i2param.pfp_ita2_btb.btb_used = 1;
-  i2param.pfp_ita2_btb.btb_tm  = 0x2;	/* capture taken-branches only */
-  i2param.pfp_ita2_btb.btb_ptm = 0x3;	/* capture regardless of tgt pred. */
-  i2param.pfp_ita2_btb.btb_ppm = 0x3;	/* capture regardless of path pred. */
-  i2param.pfp_ita2_btb.btb_brt = 0x2;	/* capture only return branches */
-  i2param.pfp_ita2_btb.btb_plm = privilege_level_mask;
-
-  iparam.pfp_dfl_plm = privilege_level_mask;
-  iparam.pfp_event_count = 1;
-  /* for system-wide monitoring we must use privileged monitors: */
-  iparam.pfp_flags = PFMLIB_PFP_SYSTEMWIDE;
-  if (pfm_find_event_byname ("BRANCH_EVENT", &iparam.pfp_events[0].event)
-      != PFMLIB_SUCCESS)
-    panic ("pfm_find_event_byname: failed to find BRANCH_EVENT\n");
-
-  memset (&oparam, 0, sizeof (oparam));
-  ret = pfm_dispatch_events (&iparam, &i2param, &oparam, NULL);
-  if (ret != PFMLIB_SUCCESS)
-    panic ("pfm_dispatch_events(): %s\n", pfm_strerror (ret));
-
-  /* Now setup the PMC and PMD descriptors: */
-
-  for (i = 0; i < oparam.pfp_pmc_count; ++i)
-    {
-      for (j = 0; j < num_pcs; ++j)
-	if (oparam.pfp_pmcs[i].reg_num == pc[j].reg_num)
-	  panic ("%s: PMC%d is already busy!\n", __FUNCTION__, pc[j].reg_num);
-      pc[num_pcs + i].reg_num = oparam.pfp_pmcs[i].reg_num;
-      pc[num_pcs + i].reg_value = oparam.pfp_pmcs[i].reg_value;
-    }
-
-  /* set the PMD reg # for BRANCH_EVENT */
-  pd[num_pds].reg_num = pc[num_pcs].reg_num;
-  branch_event_pd = pd + num_pds;
-  branch_event_pmd = pd[num_pds].reg_num;
-
-  /* Specify what happens when the BRANCH_EVENT counter wraps-around: */
-
-  pc[num_pcs].reg_smpl_pmds[0] = BTB_REGS_MASK;
-  pc[num_pcs].reg_flags |= PFM_REGFL_OVFL_NOTIFY; /* notify on BRANCH_EVENT */
-  pc[num_pcs].reg_flags |= PFM_REGFL_RANDOM;	/* randomize the interval */
-  /* clear BTB index (PMD16):  */
-  pc[num_pcs].reg_reset_pmds[0] = REG_MASK (16);
-
-  reset_value = -100000;
-  pd[num_pds].reg_value = reset_value;		/* initial value */
-  pd[num_pds].reg_long_reset = reset_value;	/* min-long-interval */
-  pd[num_pds].reg_short_reset = reset_value;	/* min-short-interval */
-  pd[num_pds].reg_random_seed = 0xc0ffee;	/* seed */
-  pd[num_pds].reg_random_mask = 0x3ff;		/* mask */
-
-  /* Define the reset value for PMD16: */
-  pd[num_pds + 1].reg_num = 16;
-  pd[num_pds + 1].reg_value = 0;
-  pd[num_pds + 1].reg_long_reset = 0;
-  pd[num_pds + 1].reg_short_reset = 0;
-
-  /* Commit the new pc/pd structures: */
-  *num_pcsp += oparam.pfp_pmc_count;
-  *num_pdsp += 2;
-}
 
 static void
 setup_code_sampling (pfarg_reg_t *pc, int *num_pcsp,
@@ -1089,24 +789,20 @@ setup_code_sampling (pfarg_reg_t *pc, int *num_pcsp,
   unsigned long reset_value;
   unsigned long i;
 
-  memset (&iparam, 0, sizeof (&iparam));
+  memset (&iparam, 0, sizeof (iparam));
 
   iparam.pfp_dfl_plm = privilege_level_mask;
   iparam.pfp_event_count = 1;
   /* for system-wide monitoring we must use privileged monitors: */
   iparam.pfp_flags = PFMLIB_PFP_SYSTEMWIDE;
+
 #if GROSS_HACK
   ++iparam.pfp_event_count;
-  if (pfm_find_event_byname ("BRANCH_EVENT", &iparam.pfp_events[0].event)
-      != PFMLIB_SUCCESS)
-    panic ("pfm_find_event_byname: failed to find BRANCH_EVENT\n");
-  if (pfm_find_event_byname (cs_event_name, &iparam.pfp_events[1].event)
-      != PFMLIB_SUCCESS)
-    panic ("pfm_find_event_byname: failed to find %s\n", cs_event_name);
+
+  iparam.pfp_events[0].event = cs_event_code;
+  iparam.pfp_events[1].event = cs_event_code;
 #else
-  if (pfm_find_event_byname (cs_event_name, &iparam.pfp_events[0].event)
-      != PFMLIB_SUCCESS)
-    panic ("pfm_find_event_byname: failed to find %s\n", cs_event_name);
+  iparam.pfp_events[0].event = cs_event_code;
 #endif
 
   memset (&oparam, 0, sizeof (oparam));
@@ -1144,7 +840,7 @@ setup_code_sampling (pfarg_reg_t *pc, int *num_pcsp,
   pc[num_pcs].reg_flags |= PFM_REGFL_OVFL_NOTIFY; /* notify on overflow */
   pc[num_pcs].reg_flags |= PFM_REGFL_RANDOM;	/* randomize the interval */
 
-  if (strcmp (cs_event_name, "CPU_CYCLES") == 0)
+  if (cs_event_code == cs_cycle_code)
     reset_value = - (long) (cycle_frequency / code_sample_rate);
   else
     reset_value = -100000;	/* your guess is as good as mine... */
@@ -1160,96 +856,9 @@ setup_code_sampling (pfarg_reg_t *pc, int *num_pcsp,
   *num_pdsp += 1;
 }
 
-static void
-setup_btb_code_sampling (pfarg_reg_t *pc, int *num_pcsp,
-			 pfarg_reg_t *pd, int *num_pdsp,
-			 unsigned int privilege_level_mask)
-{
-  int j, ret, type, num_pcs = *num_pcsp, num_pds = *num_pdsp;
-  pfmlib_ita2_input_param_t i2param;
-  pfmlib_input_param_t iparam;
-  pfmlib_output_param_t oparam;
-  unsigned long reset_value;
-  unsigned long i;
 
-  pfm_get_pmu_type (&type);
-  if (type != PFMLIB_ITANIUM2_PMU)
-    {
-      char model[80];
-      pfm_get_pmu_name (model, sizeof (model));
-      fprintf (stderr, "Interruption monitoring not supported by %s PMU; "
-	       "disabling it.\n", model);
-      return;		/* not a fatal error... */
-    }
 
-  memset (&iparam, 0, sizeof (&iparam));
-  memset (&i2param, 0, sizeof (i2param));
 
-  i2param.pfp_ita2_btb.btb_used = 1;
-  i2param.pfp_ita2_btb.btb_tm  = 0x3;	/* capture regardless whether taken */
-  i2param.pfp_ita2_btb.btb_ptm = 0x3;	/* capture regardless of tgt pred. */
-  i2param.pfp_ita2_btb.btb_ppm = 0x3;	/* capture regardless of path pred. */
-  i2param.pfp_ita2_btb.btb_brt = 0x0;	/* capture all branch-types */
-  i2param.pfp_ita2_btb.btb_plm = privilege_level_mask;
-
-  iparam.pfp_dfl_plm = privilege_level_mask;
-  iparam.pfp_event_count = 1;
-  /* for system-wide monitoring we must use privileged monitors: */
-  iparam.pfp_flags = PFMLIB_PFP_SYSTEMWIDE;
-  if (pfm_find_event_byname (cs_event_name, &iparam.pfp_events[0].event)
-      != PFMLIB_SUCCESS)
-    panic ("pfm_find_event_byname: failed to find %s\n", cs_event_name);
-
-  memset (&oparam, 0, sizeof (oparam));
-  ret = pfm_dispatch_events (&iparam, &i2param, &oparam, NULL);
-  if (ret != PFMLIB_SUCCESS)
-    panic ("pfm_dispatch_events(): %s\n", pfm_strerror (ret));
-
-  /* Now setup the PMC and PMD descriptors: */
-
-  for (i = 0; i < oparam.pfp_pmc_count; ++i)
-    {
-      for (j = 0; j < num_pcs; ++j)
-	if (oparam.pfp_pmcs[i].reg_num == pc[j].reg_num)
-	  panic ("%s: PMC%d is already busy!\n", __FUNCTION__, pc[j].reg_num);
-      pc[num_pcs + i].reg_num = oparam.pfp_pmcs[i].reg_num;
-      pc[num_pcs + i].reg_value = oparam.pfp_pmcs[i].reg_value;
-    }
-
-  /* set the PMD reg # for code-sampling event */
-  pd[num_pds].reg_num = pc[num_pcs].reg_num;
-  cs_pd = pd + num_pds;
-  cs_event_pmd = pd[num_pds].reg_num;
-
-  /* Specify what happens when the code sampling event-counter wraps-around: */
-
-  pc[num_pcs].reg_smpl_pmds[0] = BTB_REGS_MASK;
-  pc[num_pcs].reg_flags |= PFM_REGFL_OVFL_NOTIFY; /* notify on overflow */
-  pc[num_pcs].reg_flags |= PFM_REGFL_RANDOM;	/* randomize the interval */
-  /* clear BTB index (PMD16):  */
-  pc[num_pcs].reg_reset_pmds[0] = REG_MASK (16);
-
-  if (strcmp (cs_event_name, "CPU_CYCLES") == 0)
-    reset_value = - (long) (cycle_frequency / code_sample_rate);
-  else
-    reset_value = -100000;	/* your guess is as good as mine... */
-
-  pd[num_pds].reg_value = reset_value;		/* initial value */
-  pd[num_pds].reg_long_reset = reset_value;	/* min-long-interval */
-  pd[num_pds].reg_short_reset = reset_value;	/* min-short-interval */
-  pd[num_pds].reg_random_seed = 0xc0ffee;	/* seed */
-  pd[num_pds].reg_random_mask = 0x3ff;		/* mask */
-
-  /* Define the reset value for PMD16: */
-  pd[num_pds + 1].reg_num = 16;
-  pd[num_pds + 1].reg_value = 0;
-  pd[num_pds + 1].reg_long_reset = 0;
-  pd[num_pds + 1].reg_short_reset = 0;
-
-  /* Commit the new pc/pd structures: */
-  *num_pcsp += oparam.pfp_pmc_count;
-  *num_pdsp += 2;
-}
 
 unsigned long
 get_cycle_frequency (void)
@@ -1272,6 +881,59 @@ get_cycle_frequency (void)
   return mhz * 1000000 + frac;
 }
 
+static void
+init_qsys_btb_support(void)
+{
+  int pmu_type;
+  struct qsys_btb_support **q;
+
+  pfm_get_pmu_type (&pmu_type);
+  q = qsys_support_list;
+  while (*q)
+    {
+      if ((*q)->pmu_type == pmu_type)
+	goto found;
+      q++;
+  }
+  if (monitor_interruptions)
+    panic("PMU model does not support interruption monitoring, non Branch Trace Buffer\n");
+  return;
+found:
+  qsys = *q;
+}
+
+/*
+ * autodetect cycle event if not set by user
+ */
+static void
+init_default_cycle_event(void)
+{
+  size_t len;
+  
+  if (pfm_get_cycle_event(&cs_cycle_code) != PFMLIB_SUCCESS)
+    panic ("pfm_get_cycle_event: no cycle event defined\n");
+
+  if (cs_event_name == NULL)
+    {
+      pfm_get_max_event_name_len(&len);
+
+      cs_event_name = malloc(len+1);
+      if (cs_event_name == NULL)
+        panic("cannot allocate memory for cycle event name\n");
+
+      cs_event_code = cs_cycle_code;
+      if (pfm_get_event_name(cs_event_code, cs_event_name, len)
+        != PFMLIB_SUCCESS)
+        panic ("pfm_get_event_name: failed\n");
+    }
+  else
+    {
+      if (pfm_find_event_byname (cs_event_name, &cs_event_code)
+        != PFMLIB_SUCCESS)
+        panic ("pfm_find_event_byname: failed to find %s\n", cs_event_name);
+    }
+}
+
 static void
 profile_cpu (int cpu)
 {
@@ -1295,9 +957,6 @@ profile_cpu (int cpu)
     printf ("[%d] profiling on cpu %d with clock-frequency %lu Hz\n",
 	    getpid (), cpu, cycle_frequency);
 
-  if (pfm_initialize () != PFMLIB_SUCCESS)
-    panic ("Can't initialize perfmon library (pfm_initialize() failed)\n");
-
   memset (&options, 0, sizeof (options));
   options.pfm_debug = 0;	/* set to 1 for debug */
   options.pfm_verbose = 0;	/* set to 1 for debug */
@@ -1318,10 +977,20 @@ profile_cpu (int cpu)
     plm |= PFM_PLM3;
 
   if (monitor_interruptions)
-    setup_btb_code_sampling (pc, &num_pcs, pd, &num_pds, plm);
+    {
+      if(qsys)
+        {
+          qsys->setup_btb_code_sampling (pc, &num_pcs, pd, &num_pds, plm, &cs_pd);
+          cs_event_pmd = cs_pd->reg_num;
+	}
+    }
   else
     {
-      setup_call_count_sampling (pc, &num_pcs, pd, &num_pds, plm);
+      if (qsys)
+        {
+          qsys->setup_call_count_sampling (pc, &num_pcs, pd, &num_pds, plm, &branch_event_pd);
+          branch_event_pmd = branch_event_pd->reg_num;
+	}
       setup_code_sampling (pc, &num_pcs, pd, &num_pds, plm);
     }
 
@@ -1366,6 +1035,12 @@ profile_cpu (int cpu)
   sampling_buffer = context.ctx_arg.ctx_smpl_vaddr;
   sampling_fd = context.ctx_arg.ctx_fd;
 
+  if (PFM_VERSION_MAJOR(sampling_buffer->hdr_version) != PFM_VERSION_MAJOR(PFM_DEFAULT_SMPL_VERSION))
+    panic ("kernel default v%u.%u sampling format is not supported\n",
+	   PFM_VERSION_MAJOR (sampling_buffer->hdr_version),
+	   PFM_VERSION_MINOR (sampling_buffer->hdr_version));
+
+
   if (perfmonctl (sampling_fd, PFM_WRITE_PMCS, pc, num_pcs) < 0)
     panic ("perfmonctl(PFM_WRITE_PMCS) failed (%s)\n", strerror (errno));
   if (perfmonctl (sampling_fd, PFM_WRITE_PMDS, pd, num_pds) < 0)
@@ -1513,9 +1188,12 @@ main (int argc, char **argv)
   else
     prog_name = argv[0];
 
+  if (pfm_initialize () != PFMLIB_SUCCESS)
+    panic ("Cannott initialize perfmon library: maybe unsupported PMU model\n");
+
   while (1)
     {
-      ch = getopt_long (argc, argv, "c:C:e:hikm:t:uv",
+      ch = getopt_long (argc, argv, "+c:C:e:hikm:t:uv",
 			long_opts, &option_index);
       if (ch == -1)
 	break;
@@ -1591,6 +1269,10 @@ main (int argc, char **argv)
 	  break;
 	}
     }
+  init_qsys_btb_support();
+
+
+  init_default_cycle_event();
 
   if (!monitor_kernel && !monitor_user)
     monitor_kernel = monitor_user = 1;
